/***************************************************************************
                          lunacal.h  -  description
                             -------------------
    Open Source Code. See main.cpp for details.                         
    copyright            : (C) 2013 by Gerhard Holtkamp
 ***************************************************************************/

#ifndef LUNACAL_H
#define LUNACAL_H

// include files for QT
#include <QMainWindow>

class QAction;
class QMenu;

#include <QApplication>
#include <QAction>
#include <QMenuBar>
#include <QToolButton>
#include <QStatusBar>
#include <QPixmap>
#include <QPainter>
#include <QString>
#include <QKeyEvent>
#include <QLabel>
#include <QScrollArea>

#include "lunacalview.h"

class LunacalApp : public QMainWindow
{
  Q_OBJECT
  
  public:
    LunacalApp();
    ~LunacalApp();

    void initView();

    bool queryExit();
    virtual void keyPressEvent(QKeyEvent *e);

  public slots:

    void slotFileSave();
    void slotFileSaveAs();
    void slotFileQuit();
    void slotInputDate();
    void slotInputDeltaTime();
    void slotInputLocation();
    void slotCalcPhases();
    void slotCalcCalendar();
    void slotCalcEclipses();
    void slotCalcHolidays();
    void slotCalcDay();
    void slotCalcSky();
    void slotCalcYearTot();
    void slotCalcSolarPanel();
    void slotCalcSatEcl();
    void slotViewMapGrid();
    void slotViewFontSize();
    void slotViewToolBar(bool toggle);
    void slotViewStatusBar(bool toggle);

    void slotHelpAbout();
    void slotQuickHelp();

  private:

    LunacalView *view;
    QLabel *imageLabel;
    QScrollArea *sv;
   
    void createActions();
    void createMenus();
    void createToolBars();
    void createStatusBar();

    QMenu *fileMenu;
    QMenu *inputMenu;
    QMenu *calcMenu;
    QMenu *viewMenu;
    QMenu *helpMenu;

    QToolBar *fileToolBar;

    QAction *fileSave;
    QAction *fileSaveAs;
    QAction *fileQuit;
    QAction *inputDate;
    QAction *inputDeltaTime;
    QAction *inputLocation;

    QAction *calcPhases;
    QAction *calcCalendar;
    QAction *calcEclipses;
    QAction *calcHolidays;
    QAction *calcDay;
    QAction *calcSky;
    QAction *calcYearTot;
    QAction *calcSolarPanel;
    QAction *calcSatEcl;
    
    QAction *viewFontSize;
    QAction *viewMapGrid;
    QAction *viewToolBar;
    QAction *viewStatusBar;

    QAction *helpAboutApp;
    QAction *aboutQtAct;
    QAction *quickHelp;

    QAction *whts;	

};
#endif 


