/***************************************************************************
                          luncalview.cpp  -  description
                             -------------------
    Open Source Code. See main.cpp for details.
    copyright            : (C) 2012 by Holtkamp
 ***************************************************************************/
#include <QtGui>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>

using namespace std;

#include "astrolib.h"  
#include "lunacalview.h"
#include "lunacal.h"
#include "lnecldialog.h"
#include "lnsatdialog.h"
#include "nolocal.h"

int epos[GBL_ECLBUF];  // y position of eclipse info for eclipse j
int necl = 0;  // maximum number of eclipse events recorded in epos;

WoutData *outdatv = new WoutData;
WinpData *indatv = new WinpData;
int fsz = 10;  // display font size
int dsplmd = 0; // display mode. 0: no display; 1: tables; 2: eclipse text;
                //  3: eclipse map; 4 special eclipse map
bool physmap;  // true if physical map present
int mapprj;  // map projection
double centmer;  // central meridian for the map
int mapwidth, mapheight;

LunacalView::LunacalView(QWidget *parent)
 : QScrollArea(parent)
{
 QSize size;
 size.setWidth(GBL_TEXTWIDTH);
 size.setHeight(GBL_MAPWIDTH+GBL_MAPTEXTHT);
 resize(size);
 buffer = new QPixmap(size);
 buffer->fill();
 size.setWidth(GBL_MAPWIDTH);
 size.setHeight(GBL_MAPHEIGHT+GBL_MAPTEXTHT);
 buffPWM = new QPixmap(size);
 buffoot = new QPixmap(GBL_MAPWIDTH,GBL_MAPTEXTHT);
 satbuf = new QPixmap(size);
 satbuf->fill();

 // load physical world map
 if (buffPWM->load(":/lunpearth.png")) physmap = true;
 else physmap = false;

 mapwidth = 860;
 mapheight = 501;
   
}

LunacalView::~LunacalView()
{
}

void LunacalView::update()
{
  viewport()->repaint();
}

void LunacalView::setFont(int fnt)
{
 fsz = fnt;  // set display font size
}

void LunacalView::paintEvent( QPaintEvent * )
{
 QPainter painter(viewport());

 if (dsplmd == 4) painter.drawPixmap(0,0, *satbuf);
 else painter.drawPixmap(0,0, *buffer);

 painter.end();

}

void LunacalView::keyPress(int n)
{
  QString qs;
  int k;

  viewport()->clearFocus();

  if (dsplmd != 2) return;
  if (n > indatv->eclbuf.numecl) return;
  k = epos[n-1] + 2;
  showEclDetails(k);
}

void LunacalView::scrollDsp(int dx, int dy, WinpData* indata)
{
 QPoint p0(0,0);
 QPoint poffset;
 int k;

 if (dsplmd == 4)
  {
   if (dx < 0) pltStep(indata);
   else pltBack(indata); 
  }
 else
  { 
   poffset = this->mapFromParent(p0);
   k = dy - poffset.y();
   if (k > 0) k = 0;
   this->move(0,k);
  };   

}

void LunacalView::showtxt(bool tab, WoutData* outdata)
{
  char s[81];
  int i, j, x, y, dx, mxl;  
  QString qs;
  dsplmd = 1;
  
  this->unsetCursor();
   
  QFont f ("helvetica", fsz);
  buffer->fill();

  y = 0;
  
  QPainter painter;
  painter.begin(buffer);
  painter.setFont(f);
  QFontMetrics fm = painter.fontMetrics();
  dx = fm.width("M");
  painter.setPen(QColor(0,0,0));
  int maxlen = GBL_MAXCPL;

  for (i = 0; i < outdata->curlines; ++i)
   {
    y += fm.ascent();
    for (j = 0; j < maxlen; j++) s[j] = outdata->wbuf[i][j];
    qs = s;
    x = 10;
    mxl = qs.length();

    if (tab && (s[maxlen-1] == '~')) // tabular output
     {
      for (j = 0; j < mxl; j++)
       {
        qs = outdata->wbuf[i][j];
        painter.drawText( x, y, qs);
        x += dx;
       }; 
     }
    else
     {
      for (j = 0; j < mxl; j++)
       {
        qs = outdata->wbuf[i][j];
        painter.drawText( x, y, qs);
        if (outdata->wbuf[i][j] == '\t') x = 10 + j*dx;
        else x += fm.width(outdata->wbuf[i][j]);
       };
     };
   };

   painter.end();

   update();
}

void LunacalView::showEcltxt(WinpData* indata)
{
  char s[81];
  char n[3];
  int i, j, x, y, dx, mxl, py;
  QString qs, s2;

  dsplmd = 2;
  this->unsetCursor();
  
  indatv->day = indata->day;
  indatv->month = indata->month;
  indatv->year = indata->year;
  indatv->time = 0;
  indatv->tzone = indata->tzone;
  indatv->del_tdut = indata->del_tdut;
  indatv->tmfirst = 1;
  indatv->geolat = indata->geolat;
  indatv->geolong = indata->geolong;
  indatv->geoheight = indata->geoheight;
  for(j=0; j<31; j++) indatv->geoname[j] = indata->geoname[j];
  indatv->dmsflg = indata->dmsflg;
  indatv->filsflg = 0;

  outdatv->curlines = 0;
  moonph(indatv, outdatv);
  outdatv->curlines = 0;
  PutEclLst(indatv, outdatv);
  
  QFont f ("helvetica", fsz);
  resize(820,1600);
  buffer->fill();

  y = 0;

  QPainter painter;
  painter.begin(buffer);
  painter.initFrom(viewport());
  painter.setFont(f);
  QFontMetrics fm = painter.fontMetrics();
  dx = fm.width("M");
  painter.setPen(QColor(0,0,0));
  int maxlen = GBL_MAXCPL;

  necl = 0;
  
  for (i = 0; i < outdatv->curlines; ++i)
   {
    y += fm.ascent();
    for (j = 0; j < maxlen; j++) s[j] = outdatv->wbuf[i][j];
    qs = s;
    x = 10;
    mxl = qs.length();

    // record y position of eclipse info
    if (s[1] != ' ')  // this line contains eclipse info
     {
      n[0] = s[0];
      n[1] = s[1];
      n[2] = '\0';
      s2 = n;
      py = s2.toInt();
      if (py <= GBL_ECLBUF)
       {
        epos[py-1] = y - fm.ascent() - 2;
        necl++;
       }; 
     };

    if (s[maxlen-1] == '~') // tabular output
     {
      for (j = 0; j < mxl; j++)
       {
        qs = outdatv->wbuf[i][j];
        painter.drawText( x, y, qs);
        x += dx;
       };
     }
    else
     {
      for (j = 0; j < mxl; j++)
       {
        qs = outdatv->wbuf[i][j];
        painter.drawText( x, y, qs);
        if (outdatv->wbuf[i][j] == '\t') x = 10 + j*dx;
        else x += fm.width(outdatv->wbuf[i][j]);
       };
     };
   };

   painter.end();

   update();

}

void LunacalView::showEclDetails( int y )
 {
  int k, j;
  char otxt[255];
	char oldtxt[11];
  QString s, s2;

  k = -1;
  if (necl > 0)
   {
    for (j=0; j<necl; j++)
     {
      if (y > epos[j]) k = j;
     };
   };

  if (k < 0) return;

  indatv->eclbuf.eclselect = k+1;

  if (indatv->eclbuf.phase[k] > 0) showEclMap();

  sprintf(otxt,"+++> Timezone %g <+++\tYear: %4i " ,indatv->tzone, indatv->year);
  decdot(otxt);
  if (indatv->eclbuf.phase[indatv->eclbuf.eclselect-1] > 3)
   {
    if(indatv->dmsflg) strcat (otxt,"\tLat./Long. in Degrees.Minutes");
    else strcat (otxt,"\tLat./Long. in Decimal Degrees");
   };
  strcat(otxt,"\n\n");

  s = otxt;
  strcpy (oldtxt,"**********");
  indatv->eclbuf.ecloutbn = 1;
  do
   {
    EclDetails(indatv, otxt);
    if (strncmp(oldtxt,otxt,10) != 0)  // make sure lines don't appear double
     {
      s2 = otxt;
      s += s2;
      strncpy(oldtxt,otxt,10);
     }
    if (indatv->eclbuf.ecloutbn > 0) indatv->eclbuf.ecloutbn++;
   } while (indatv->eclbuf.ecloutbn > 0 );

  sprintf(otxt,
          "\n\n\n+++> tdt - ut (in sec): %g <+++" ,indatv->del_tdut);
  decdot0(otxt);
  s2 = otxt;
  s += s2;

  LnEclDialog *dlg = new LnEclDialog();
  dlg->getParams(s, fsz);
  dlg->exec();
 }

void LunacalView::mousePressEvent( QMouseEvent *e )
{
 int x, y, stpsav;

 x = e->x();
 y = e->y();

 if (dsplmd == 2)
  {
   showEclDetails(y);
   return;
  };  
 if (dsplmd == 3)
  {
   stpsav = indatv->eclbuf.cstep; 
   indatv->eclbuf.cstep = 60;
   getMapPos(x, y);
   indatv->eclbuf.cstep = stpsav;
  }; 
 
 this->setFocus();
 
}

void LunacalView::setMap(int mgrid, int mgrey, int mstep, int mpj, double mcentmer)
 {
  // set eclipse display parameters
  indatv->eclbuf.emagrid = mgrid;
  indatv->eclbuf.emagrey = mgrey;
  indatv->eclbuf.cstep = mstep;
  mapprj = mpj,
  centmer = mcentmer;
 }
 
void LunacalView::pltStep(WinpData* indata)
{
 indata->mjd += indata->tstep/86400.0;
 pltCurrent(indata);
}

void LunacalView::pltBack(WinpData* indata)
{
 indata->mjd -= indata->tstep/86400.0;
 pltCurrent(indata);
}

void LunacalView::pltCurrent(WinpData* indata)
{
  const int tc5 = 4; // = 4 for Qt4 and = 5 for Qt3  ###
  int dd, mm, yy, hh, mn, ss;
  int et1, et2, x, y, clr, clg, clb;
  double lat1, lng1, lat2, lng2;
  double hour, sec;
  char dts[31];
  char otxt[99];
  QString qs;
  QPoint p;

  caldat(indata->mjd, dd, mm, yy, hour);
  dms (hour, hh, mn, sec);
  ss = int(sec);

  /* store data in their positions */
  GetMonth (mm, dts);
  dts[3]=' ';
  sprintf(&dts[4],"%2i  %2i:%02i:%02i", dd, hh, mn, ss);
  dts[16]='\0';
  sprintf(otxt, "%4i ", yy);
  strcat (otxt, dts);
  sprintf(dts, " UT   h:%8.1lf km ", indata->eclheight);
  decdot(dts);
  strcat (otxt, dts);

  buffoot->fill(QColor(0,0,255));

  QPainter painter;
  painter.begin(satbuf);
  painter.drawPixmap(0,0,*buffer);
  painter.drawPixmap(0,362,*buffoot);

  SatEcl(indata->mjd, indata->del_tdut, indata->eclheight, indata->geoid,
              et1, lat1, lng1, et2, lat2, lng2);
  clg = 0;                      
  if (et1 > 0)  // front surface
   {
    if (et1 == 1)
     {          // colors for total eclipse
      clr = 0;
      clb = 255;
      strcat (otxt, "   tot.");
     }
    else
     {          // colors for annular eclipse
      clr = 255;
      clb = 0;
      strcat (otxt, "   ann.");
     };
    painter.setPen(QColor (clr, clg, clb));
    p = PxCoord(lat1,lng1);
    x = p.x();
    y = p.y();
    x -= 2;
    y -= 2;
    painter.drawEllipse(x, y, tc5, tc5);
    painter.drawPoint(PxCoord(lat1, lng1));
    sprintf(dts, "%7.2lf %8.2lf", lat1, lng1);
    decdot0(dts);
    strcat(otxt, dts);
   }; 

  if (et2 > 0)  // back surface
   {
    if (et2 == 1)
     {     // colors for total eclipse
      clr = 0;
      clb = 255;
      strcat (otxt, "   tot.");
     }
    else
     {       // colors for annular eclipse
      clr = 255;
      clb = 0;
      strcat (otxt, "   ann.");
     };
    painter.setPen(QColor (clr, clg, clb));
    p = PxCoord(lat2,lng2);
    x = p.x();
    y = p.y();
    x -= 2;
    y -= 2;
    painter.drawEllipse(x, y, tc5, tc5);
    painter.drawPoint(PxCoord(lat2, lng2));
    sprintf(dts, "%7.2lf %8.2lf", lat2, lng2);
    decdot0(dts);
    strcat(otxt, dts);
   };

  qs = otxt;
  painter.setPen(QColor(255,255,0));
  painter.drawText(0, 380, qs);

  painter.end();
  
  update();
}

void LunacalView::showSatEcl(WinpData* indata)
{
  int wdt, hgt;
  double hour, jd, jdf, sec, height, tstep;
  double lat1, lng1, lat2, lng2;
  int dd, mm, yy, hh, mn, ss;
  int clr, clg, clb, et1, et2;
  bool geoid;

  caldat(indata->mjd, dd, mm, yy, hour);
  dms (hour, hh, mn, sec);
  ss = int(sec);
  geoid = indata->geoid;
  height = indata->eclheight;
  tstep = indata->tstep;

  LnSatDialog *dlg = new LnSatDialog();
  dlg->setParams(geoid, height, tstep, yy, mm, dd, hh, mn, ss);
  if (dlg->exec())
   {
    dlg->getParams(geoid, height, tstep, yy, mm, dd, hh, mn, ss);
    indata->geoid = geoid;
    indata->eclheight = height;
    indata->tstep = tstep;    
    hour = ddd(hh, mn, double(ss));
    jd = mjd(dd, mm, yy, hour);
    indata->mjd = jd;
   }
  else return;  
      
  dsplmd = 4;
  viewport()->setCursor( Qt::CrossCursor );
  
  resize(buffer->width(), buffer->height());
  if (indata->eclbuf.emagrey) buffer->fill( QColor(200,200,200));
  else buffer->fill( QColor(255,255,255));

  QPainter painter;
  painter.begin(buffer);

  wdt = GBL_MAPWIDTH;
  hgt = GBL_MAPHEIGHT;

  // get origin and scale factors
  origx = wdt / 2;
  origy = hgt / 2;
  if ((mapprj == 1) || (mapprj == 2))
   {
    xscale = double(origy) / 1.678;
    yscale = xscale;
   }
  else
   { 
    xscale = double(origx) / 180.0;
    yscale = double(origy) / 90.0;
   }; 

  PlotWorld(painter, indata);

  jdf = jd + 2.0;
  jd -= 2.0;
  clg = 0;
  do
   {
    SatEcl(jd, indata->del_tdut, indata->eclheight, indata->geoid,
               et1, lat1, lng1, et2, lat2, lng2);
    if (et1 > 0)  // front surface
     {                    
      if (et1 == 1)
       {     // colors for total eclipse
        clr = 0;
        clb = 255;
       }
      else
       {     // colors for annular eclipse
        clr = 255;
        clb = 0;
       };

      painter.setPen(QColor (clr, clg, clb));
      painter.drawPoint(PxCoord(lat1, lng1));
     };
      
    if (et2 > 0)  // back surface
     {
      if (et2 == 1)
       {     // colors for total eclipse
        clr = 0;
        clb = 255;
       }
      else
       {       // colors for annular eclipse
        clr = 255;
        clb = 0;
       };

      painter.setPen(QColor (clr, clg, clb));
      painter.drawPoint(PxCoord(lat2, lng2));
     };

    jd += 0.000694444;  // add 1 minute  
   } while (jd < jdf);

  pltCurrent(indata);
}

void LunacalView::showEclMap()
{
  int wdt, hgt, j, p;
  char jtxt[21];
  char otxt[51];
  QString qs;

  buffoot->fill(QColor(0,0,255));

  dsplmd = 3;
  viewport()->setCursor( Qt::CrossCursor );

  resize(buffer->width(), buffer->height());
  if (indatv->eclbuf.emagrey) buffer->fill( QColor(200,200,200));
  else buffer->fill( QColor(255,255,255));

  QPainter painter;
  painter.begin(buffer);
  painter.drawPixmap(0,362,*buffoot);

  wdt = GBL_MAPWIDTH;
  hgt = GBL_MAPHEIGHT;

  // get origin and scale factors
  origx = wdt / 2;
  origy = hgt / 2;
  if ((mapprj == 1) || (mapprj == 2))
   {
    xscale = double(origy) / 1.678;
    yscale = xscale;
   }
  else
   { 
    xscale = double(origx) / 180.0;
    yscale = double(origy) / 90.0;
   }; 

  PlotWorld(painter, indatv);
  PltCenLin(painter, indatv);
  PltEclBound(painter, indatv);
  PltMaxPos(painter, indatv);

  // draw footer
  sprintf(otxt,"  %4i  ",indatv->year);
  j = indatv->eclbuf.eclselect-1;
  GetEclTxt(indatv, j, jtxt);
  for (p=0; p<10; p++) jtxt[p] = jtxt[p+4];
  jtxt[7] = ' ';
  jtxt[8] = '\0';

  strcat(otxt,jtxt);
  p = indatv->eclbuf.phase[j];
  switch (p)
   {
    case 4: strcat(otxt,"  Annular");
            break;
    case 5: strcat(otxt,"  Total");
            break;
    case 6: strcat(otxt,"  Annular/Total");
            break;
   };
  strcat(otxt," Eclipse of the Sun");
  qs = otxt;
  painter.setPen(QColor(255,255,0));
  painter.drawText(0, 380, qs);
  painter.end();

  update();
}

void LunacalView::getMapPos(int x, int y)
 {
  // display eclipse details from clicked map position
  int j, k;
  double latsav, lngsav, hgtsav, lat, lng;
  char nsav[31];
  QPoint p;

  latsav = indatv->geolat;
  lngsav = indatv->geolong;
  hgtsav = indatv->geoheight;
  for (j=0; j<31; j++) nsav[j] = indatv->geoname[j];

  p.setX(x);
  p.setY(y);
  CoordPx (p, lat, lng);  
  indatv->geolat = lat;
  indatv->geolong = lng;
  indatv->geoheight = 0;
  sprintf(indatv->geoname, "clicked position");
  j = indatv->eclbuf.eclselect - 1;
  k = epos[j]+1;
  showEclDetails(k);  
  indatv->geolat = latsav;
  indatv->geolong = lngsav;
  indatv->geoheight = hgtsav;
  for (j=0; j<31; j++) indatv->geoname[j] = nsav[j];
 };  

void LunacalView::PlotWorld(QPainter &WorldMap, WinpData* indata)
 {
  /* Draw the grid of coordinates and the outlines of continents */
  int i, j;
  int wdt, hgt;
  double s, c2;
  QPoint p, p0, p2;

  wdt = GBL_MAPWIDTH;
  hgt = GBL_MAPHEIGHT;

  // draw the continents
  MapLand(WorldMap);

  // draw 10 deg lines
  WorldMap.setPen(QColor( 130, 160, 180 ));

  if ((mapprj == 0) || (mapprj == 3))
   {
    if (indata->eclbuf.emagrid == 2)
     {
      for (j= -170; j<=180; j+=10)
       {
        s = double(j);
        p = PxCoord(90.0,s);
        i = p.x();
        WorldMap.drawLine(i,0,i,hgt);
       };
      for (j= -80; j<90; j+=10)
       {
        s = double(j);
        p = PxCoord(s,-180.0);
        i = p.y();
        WorldMap.drawLine(0,i,wdt,i);
       };
      WorldMap.setPen(QColor( 30, 240, 120 ));
     };

    // draw 30 deg lines
    if (indata->eclbuf.emagrid > 0)
     {
      for (j= -150; j<=180; j+=30)
       {
        s = double(j);
        p = PxCoord(90.0,s);
        i = p.x();
        WorldMap.drawLine(i,0,i,hgt);
       };
      for (j= -60; j<90; j+=30)
       {
        s = double(j);
        p = PxCoord(s,-180.0);
        i = p.y();
        WorldMap.drawLine(0,i,wdt,i);
       };
     };
   }
  else
   {
    // draw 30 deg lines stereographic
    if (indata->eclbuf.emagrid > 0)
     {
      if (mapprj == 1)
       {
        p0 = PxCoord(90.0,0);
        c2 = 10.0;
       } 
      else
       {
        p0 = PxCoord(-90.0,0);
        c2 = -10.0;
       }; 
      for (j= -180; j<180; j+=30)
       {
        s = double(j);
        p = PxCoord(c2,s);
        WorldMap.drawLine(p0,p);
        // the following is to avoid plotting out of the field
        if (mapprj == 1)
         {
          for (i=9; i>= 0; i--)
           {
            p2 = p;
            p = PxCoord(double(i),s);
            if (p.y() < mapwidth) WorldMap.drawLine(p2,p);
           };  
         }
        else
         {
          for (i=-9; i<= 0; i++)
           {
            p2 = p;
            p = PxCoord(double(i),s);
            if (p.y() < mapwidth) WorldMap.drawLine(p2,p);
           };
         };     
       };
  
      if (mapprj == 1) s = 60.0;
      else s = -60.0;
      p0 = PxCoord(s,-180.0);
      for (j= -175; j<=180; j+=5)
       {
        p = PxCoord(s,double(j));
        WorldMap.drawLine(p0,p);
        p0 = p;
       };
      if (mapprj == 1) s = 30.0;
      else s = -30.0;
      p0 = PxCoord(s,-180.0);
      for (j= -175; j<=180; j+=5)
       {
        p = PxCoord(s,double(j));
        WorldMap.drawLine(p0,p);
        p0 = p;
       };
     };
   };        

  if (indata->eclbuf.emagrid == 0)  // right border for map
   {
    wdt = wdt - 1;
    WorldMap.drawLine(wdt,0,wdt,hgt);
   };
 }

QPoint LunacalView::PxCoord (double lat, double lng)
 {
  /* get pixel coordinates for the World Map from the
     geographic latitude lat and the longitude lng
     (both in decimal degrees)

     RETURN: the pixel coordinates as a point
  */

  const double dgrad = M_PI / 180.0;
  QPoint p;
  double x, y, s;
  int k;

  if ((mapprj == 0) || (mapprj == 3))  // normal world map
   {  
    // latitude
    s = -lat * yscale;
    if (s < 0) s -= 0.5;
    else s += 0.5;
    k = int(s);
    k += origy;
    p.setY(k);

    if (centmer != 0)
     {
      lng -= centmer;
      if (lng > 180.0) lng -= 360.0;
      if (lng < -180.0) lng += 360.0;
     }; 

    // longitude
    if (lng > 180.0) lng -= 360.0;
    if (lng < -180.0) lng += 360.0;
    s = lng * xscale;
    if (s < 0) s -= 0.5;
    else s += 0.5;
    k = int(s);
    k += origx;
    p.setX(k);
   }
  else  // stereographic projection
   { 
    if (centmer != 0)
     {
      lng -= centmer;
      if (lng > 180.0) lng -= 360.0;
      if (lng < -180.0) lng += 360.0;
     };

    if (mapprj == 2) // southern hemisphere
     {
      lat = -lat;
      lng = -lng;
     };   

    if ((lat < 0) || (lat >90.1))
     {
      p.setX(mapwidth + 3);
      p.setY(mapheight + 3);
      return p;
     };
    y = lat * dgrad;
    x = lng * dgrad;
    s = 2.0 * cos(y) / (1.0 + sin(y));
    y = s * cos(x);
    x = -s * sin(x);

    // latitude
    s = -y * yscale;
    if (s < 0) s -= 0.5;
    else s += 0.5;
    k = int(s);
    k += origy;
    p.setY(k);

    // longitude
    s = x * xscale;
    if (s < 0) s -= 0.5;
    else s += 0.5;
    k = int(s);
    k += origx;
    p.setX(k);

    if (p.x() > GBL_MAPWIDTH) p.setX(mapwidth + 3);
    if (p.y() > GBL_MAPHEIGHT)
     {
      p.setY(mapwidth + 3);
      p.setX(mapwidth + 3);
     };   
   }; 

  return p;
 }

void LunacalView::CoordPx (QPoint& p, double& lat, double& lng)
 {
  /* get coordinates of geographic latitude lat and
     longitude lng (both in decimal degrees)
     from pixel p of the World Map
  */

  const double dgrad = M_PI / 180.0;
  double s, s2, az1, el1;
  int k;

  // latitude
  k = p.y() - origy;
  s = k;
  lat = -s / yscale;

  // longitude
  k = p.x() - origx;
  s = k;
  lng = s / xscale;

  if ((mapprj == 1) || (mapprj == 2)) // stereographic projection
   {  
    if ((lng == 0) && (lat == 0))
     {
      az1 = 0;
      el1 = 90;
     }
    else
     {
      lng = -lng;
      s = atan2(lng, lat);
      az1 = s / dgrad;
      s2 = sin(s);
      if (fabs(s2) > 1e-5) s = s2;
      else
       {
        s = cos(s);
        lng = lat;
       };
      s = 0.5 * lng / s;
      lat = (s*s - 1.0) / (s*s + 1.0);
      s = s*s / (s*s + 1.0);
      lat = s*s - lat;
      if (lat < 0) el1 = -100.0;
      else
       {
        lng = -s + sqrt(lat);
        lng = asin(lng);
        el1 = lng / dgrad;
        if (el1 < 0) el1 = -100.0;
       };
     };
    lng = az1;
    lat = el1;

    if ((mapprj == 2) && (lat != -100.0))
     {
      lng = -lng;
      lat = -lat; 
     };    
    if (lng > 180.0) lng -= 360.0;
    if (lng < -180.0) lng += 360.0;
   };

  if (centmer != 0)
   {
    lng += centmer;
    if (lng > 180.0) lng -= 360.0;
    if (lng < -180.0) lng += 360.0;
   };

 }

void LunacalView::PltCenLin (QPainter &WorldMap, WinpData* indata)
 {
  // plot center line
  char otxt[255];
  char oldtxt[11];
  double lat, lng;
  int cstsav, clr, clg, clb ;

  // use 1 min time steps
  cstsav = indata->eclbuf.cstep;
  indata->eclbuf.cstep = 1;

  strcpy (oldtxt,"**********");
  indata->eclbuf.ecloutbn = 2;
  do
   {
    EclDetails(indata, otxt);
    if (strncmp(oldtxt,otxt,10) != 0)  // make sure lines don't appear double
     {
      if (indata->eclbuf.cphs > 3)
       {
        lat = indata->eclbuf.clat;
        lng = indata->eclbuf.clng;
        if (indata->eclbuf.cphs > 4)
         {           // colors for total eclipse
          clr = 0;
          clg = 0;
          clb = 255;
         }
        else
         {          // colors for annular eclipse
          clr = 255;
          clg = 0;
          clb = 0;
         };

        WorldMap.setPen(QColor (clr, clg, clb));
        WorldMap.drawPoint(PxCoord(lat, lng));
       };
      strncpy(oldtxt,otxt,10);
     }
    if (indata->eclbuf.ecloutbn > 0) indata->eclbuf.ecloutbn++;
   } while (indata->eclbuf.ecloutbn > 0 );

  // get max eclipse if not central
  if (indata->eclbuf.phase[(indata->eclbuf.eclselect-1)] < 4) GetMaxPos(indata);
   
  // restore old step width
  indata->eclbuf.cstep = cstsav;
 }

void LunacalView::PltMaxPos (QPainter &WorldMap, WinpData* indata)
 {
  // plot position of maximum eclipse

  double lat, lng;
  int x, y;
  QPoint p;

  // mark maximum eclipse
  lat = indata->eclbuf.cmxlat;
  lng = indata->eclbuf.cmxlng;
  WorldMap.setPen( QColor( 0, 255, 0 ));
  p = PxCoord(lat,lng);
  x = p.x();
  y = p.y();
  x -= 2;
  y -= 2;
  WorldMap.drawEllipse(x, y, 5, 5);
  WorldMap.setPen(QColor (0, 255, 0));
  WorldMap.drawPoint(PxCoord(lat, lng));
 }

void LunacalView::PltEclBound (QPainter &WorldMap, WinpData* indata)
 {
  // Plot the northern and southern boundaries of the eclipse
  // as well as the Sunrise and Sunset boundaries.

  double t, dt, lat, lng, lat2, lng2;
  double tj[6], jd1sav, jd2sav;
  double dpd, dpb;
  int clr, clg, clb, j ;
  Vec3 ubm, ube, udm, ude, lbe, lde;

  // set color to black
  clr = 0;
  clg = 0;
  clb = 0;

  // partition the time into 5 intervals
  jd1sav = indata->eclbuf.jdstart;
  jd2sav = indata->eclbuf.jdstop;
  tj[0] = indata->eclbuf.jdstart - indata->tzone/24.0;
  tj[5] = indata->eclbuf.jdstop - indata->tzone/24.0;
  dt = (tj[5] - tj[0]) / 5.0;
  for (j = 1; j < 5; j++) tj[j] = tj[j-1] + dt;

  // North South Boundaries
  for (j = 0; j < 5; j++)
   {
    // initialize
    indata->eclbuf.jdstart = tj[j];
    indata->eclbuf.jdstop = tj[j+1];
    InitBound(indata, ubm, ube, udm, ude, lbe, lde);
    t = indata->eclbuf.jdstart;
    if ((j < 1) || (j > 3)) dt = 1.0;
    else dt = 2.5;
    dt = dt / (24.0*60.0);  // 1 min

    do
     {
      GNSBound(indata, t, ubm, ube, udm, ude, lat, lng);
      if (lat <= 90.0)
       {
        WorldMap.setPen(QColor (clr, clg, clb));
        WorldMap.drawPoint(PxCoord(lat, lng));
       }; 
      GNSBound(indata, t, ubm, lbe, udm, lde, lat, lng);
      if (lat <= 90.0)
       {
        WorldMap.setPen(QColor (clr, clg, clb));
        WorldMap.drawPoint(PxCoord(lat, lng));
       };
      t += dt;
     } while (t < indata->eclbuf.jdstop);
   };

  // Sunrise / Sunset Boundaries
  dt = 1.0;
  dt = dt / (24.0*60.0);  // 1 min
  for (j = 0; j < 5; j++)
   {
    // initialize
    indata->eclbuf.jdstart = tj[j];
    indata->eclbuf.jdstop = tj[j+1];
    InRSBound(indata, ubm, ube, udm, ude, dpb, dpd);
    t = indata->eclbuf.jdstart;

    do
     {
      GRSBound(indata, t, ubm, ube, udm, ude, dpb, dpd, lat, lng, lat2, lng2);
      if (lat <= 90.0)
       {
        WorldMap.setPen(QColor (200, 0, 150));
        WorldMap.drawPoint(PxCoord(lat, lng));
       }; 
      if (lat2 <= 90.0)
       {
        WorldMap.setPen(QColor (200, 0, 150));
        WorldMap.drawPoint(PxCoord(lat2,lng2));
       }; 
      t += dt;
     } while (t < indata->eclbuf.jdstop);
   };

  // restore the start and stop times
  indata->eclbuf.jdstart = jd1sav;
  indata->eclbuf.jdstop = jd2sav;
 }
  
void LunacalView::MapLand(QPainter &WorldMap)
 {
  /* Draw the continents and major islands
     The geographic coordinates are defined in an array geo
     with odd index latitude and the next even indes longitude.
     The coordinates are stored as integers in tens of degrees
     (thus 35.2 degrees will be stored as 352).
     West and South are negative.
  */
  const int gx = 3598;
  int j, k;
  int lon, ofz;
  double lat, lng;
  QPoint p1, p2;

  int geo[gx] = {
	  // Australia  206
	  1111, 0,
	  -200, 1200,  -200, 1183,  -204, 1180,  -210, 1175,
	  -210, 1175,  -207, 1169,  -218, 1160,  -222, 1145,
	  -226, 1140,  -226, 1137,  -222, 1136,  -227, 1130,
	  -234, 1134,  -241, 1128,  -260, 1140,  -267, 1135,
	  -260, 1128,  -270, 1130,  -265, 1125,  -275, 1140,
	  -280, 1140,  -295, 1150,  -305, 1150,  -325, 1158,
	  -335, 1155,  -337, 1150,  -342, 1150,  -348, 1159,
	  -350, 1165,  -350, 1177,  -337, 1201,  -333, 1224,
	  -335, 1226,  -335, 1238,  -328, 1240,  -322, 1265,
	  -317, 1300,  -315, 1317,  -320, 1338,  -350, 1360,
	  -325, 1375,  -353, 1370,  -351, 1377,  -332, 1376,
	  -348, 1378,  -360, 1367,  -357, 1378,  -375, 1397,
	  -383, 1405,  -388, 1438,  -377, 1448,  -395, 1462,
	  -375, 1477,  -375, 1495,  -372, 1500,  -365, 1500,
	  -350, 1510,  -338, 1523,  -328, 1525,  -300, 1530,
	  -277, 1538,  -266, 1528,  -245, 1527,  -252, 1525,
	  -236, 1509,  -225, 1505,  -225, 1495,  -215, 1487,
	  -182, 1466,  -170, 1460,  -165, 1456,  -150, 1452,
	  -142, 1446,  -145, 1435,  -125, 1430,  -110, 1423,
	  -113, 1420,  -137, 1418,  -150, 1420,  -176, 1410,
	  -176, 1400,  -147, 1355,  -123, 1372,  -120, 1362,
	  -123, 1350,  -117, 1321,  -124, 1325,  -122, 1312,
	  -125, 1306,  -140, 1292,  -150, 1295,  -145, 1277,
	  -133, 1270,  -146, 1253,  -153, 1245,  -166, 1243,
	  -165, 1232,  -178, 1230,  -168, 1225,  -175, 1223,
	  -180, 1220,  -200, 1200,
	  // Tasmania   20
	  1111, 0,
	  -405, 1445,  -440, 1460,  -442, 1470, -438, 1473,
	  -430, 1475,  -425, 1479,  -405, 1475, -415, 1469,
	  -405, 1445,
	  // New Zealand  174
	  1111, 0,    // North Island
	  -344, 1726,  -352, 1731,  -353, 1731,  -369, 1743,
	  -377, 1746,  -381, 1747,  -382, 1745,  -383, 1746,
	  -390, 1744,  -392, 1739,  -394, 1737,  -396, 1740,
	  -397, 1744,  -399, 1747,  -399, 1750,  -404, 1752,
	  -411, 1747,  -413, 1746,  -416, 1748,  -415, 1753,
	  -417, 1753,  -413, 1760,  -406, 1764,  -405, 1766,
	  -396, 1771,  -395, 1768,  -391, 1772,  -391, 1778,
	  -393, 1778,  -391, 1780,  -390, 1779,  -387, 1779,
	  -386, 1782,  -384, 1784,  -380, 1784,  -376, 1786,
	  -375, 1780,  -380, 1773,  -376, 1762,  -365, 1754,
	  -372, 1754,  -372, 1753,  -369, 1752,  -367, 1747,
	  -363, 1748,  -358, 1745,  -358, 1746,  -352, 1743,
	  -350, 1738,  -350, 1734,  -345, 1730,  -344, 1726,
	  1111, 0,     // South Island
	  -405, 1726,  -409, 1721,  -414, 1721,  -418, 1714,
	  -425, 1712,  -430, 1706,  -440, 1687,  -440, 1684,
	  -457, 1665,  -460, 1665,  -463, 1666,  -463, 1676,
	  -466, 1683,  -465, 1687,  -467, 1690,  -465, 1698,
	  -459, 1704,  -458, 1708,  -456, 1706,  -449, 1712,
	  -444, 1712,  -438, 1727,  -438, 1732,  -436, 1730,
	  -436, 1726,  -429, 1734,  -417, 1743,  -414, 1741,
	  -410, 1742,  -407, 1739,  -413, 1732,  -407, 1730,
	  -405, 1726,
	  // Africa   250  (cum 650)
	  1111, 0,
	  364, -55,  360, -60,  340, -70,  332, -85,  318, -97,
	  305, -95,  285, -110,  280, -130,  270, -140,  245, -150,
	  240, -160,  210, -170,  200, -165,  192, -168,  158, -170,
	  142, -175,  134, -170,  122, -165,  80, -130,  73, -125,
	  42, -77,  50, -45,  45, -26,  60, 35,  55, 42,
	  45, 50,  45, 90,  40, 90,  35, 97,  15, 92,
	  8, 95,  2, 92,  -4, 90,  -58, 123,  -62, 125,
	  -92, 130,  -92, 127,  -128, 131,  -140, 125,  -158, 118,
	  -180, 116,  -227, 145,  -245, 145,  -250, 150,  -260, 155,
	  -275, 160,  -285, 165,  -323, 180,  -330, 175,  -345, 185, // cape argulhas
	  -350, 200,  -345, 208,  -343, 215,  -338, 222,  -340, 250,
	  -335, 265,  -311, 302,  -290, 318,  -285, 323,  -255, 330,
	  -245, 360,  -222, 355,  -200, 350,  -175, 370,  -160, 400,
	  -150, 405,  -103, 404,  -76, 386,  -65, 390,  -55, 382,
	  -45, 390,  -44, 396,  -25, 402,  -25, 408,  20, 445,
	  25, 460,  50, 482,  91, 506,  118, 510,  106, 440,
	  112, 425,  116, 420,  120, 428,  130, 422,  152, 392,
	  176, 383,  183, 374,  223, 370,  231, 352,  240, 355,
	  297, 322,   1111, 0,  // suez  (186)
	  318, 320,  322, 307,  310, 290,  318, 258,  316, 248,
	  320, 248,  325, 227,  328, 225,  331, 220,  320, 200,
	  306, 200,  302, 190,  312, 170,  317, 155,  322, 152,
	  330, 136,  328, 120,  340, 100,  352, 112,  360, 105,
	  370, 110,  368, 104,  372, 100,  370, 88,  372, 62,
	  367, 52,  370, 40,  365, 20,  350, -25,  350, -38,
	  355, -55, 364, -55,    // (250)
	  // Madagascar  48  (cum 698)
	  1111, 0,
	  -200, 445,  -217, 440,  -225, 435,  -234, 438,  -250, 440,
	  -257, 445,  -263, 450,  -255, 472,  -177, 492,  -172, 490,
	  -160, 496,  -155, 496,  -158, 504,  -152, 507,  -128, 500,
	  -120, 490,  -123, 485,  -132, 480,  -137, 475,  -150, 468,
	  -160, 440,  -170, 435,  -200, 445,
	  // Lake Victoria  16
	  1111, 0,
	  0, 330,  -15, 325,  -35, 328,  -35, 340,  -5, 345,
	  3, 340,  0, 330,
	  // Lake Tanganyika  16
	  1111, 0,
	  -40, 292,  -55, 292,  -75, 306,  -80, 315,  -60, 300,
	  -50, 297,  -40, 292,
	  // Lake Malawi   18  (cum 748)
	  1111, 0,
	  -96, 340,  -148, 346,  -145, 350, -123, 347,  -120, 344,
	  -117, 350,  -100, 345, -96, 340,
	  // South America  340  (cum 1088)
	  1111, 0,
	  90, -796,  85, -790,  77, -775,  72, -772,  60, -772,
	  58, -775,  42, -772,  40, -770,  28, -777,  25, -780,
	  20, -777,  12, -805,  -5, -810,  -10, -810,  -10, -816,
	  -24, -815,  -28, -804,  -25, -795,  -30, -800,  -44, -820,
	  -60, -820,  -70, -800,  -134, -763,  -142, -763,  -152, -752,
	  -174, -715,  -185, -705,  -215, -701,  -232, -705,  -232, -708,
	  -235, -708,  -235, -705,  -245, -708,  -254, -708,  -257, -710,
	  -270, -710,  -270, -713,  -275, -710,  -285, -715,  -293, -715,
	  -300, -715,  -303, -717,  -325, -715,  -370, -732,  -372, -735,
	  -395, -733,  -400, -737,  -410, -740,  -417, -736,  -418, -740,
	  // isla chiloe
	  -434, -745,  -435, -736,  -418, -732,  -420, -725,  -438, -732,
	  -437, -743,  -455, -745,  -468, -755,  -470, -740,  -480, -755,
	  -508, -755,  -505, -750,  -515, -754,  -513, -742,  -520, -750,
	  -530, -730,  -528, -748,  -540, -735,  -545, -720,  -552, -707,
	  -550, -700,  -553, -702,  -558, -681,  -552, -685,  -553, -672,
	  -550, -670,  -549, -655,  -546, -650,  -546, -660,  -540, -675,
	  -526, -687,  -527, -695,  -522, -693,  -523, -685,  -510, -692,
	  -505, -690,  -500, -678,  -492, -675,  -480, -658,  -470, -658,
	  -470, -669,  -460, -675,  -452, -670,  -450, -655,  -435, -652,
	  -427, -635,  -421, -635,  -422, -650,  -407, -650,  -410, -628,
	  -406, -623,  -389, -624,  -389, -590,  -380, -576,  -364, -567,
	  -360, -570,  -352, -570,  -345, -588,  -342, -585,  -351, -562,
	  -352, -548,  -345, -537,  -286, -488,  -274, -484,  -272, -486,
	  -258, -485,  -240, -463,  -230, -430,  -230, -420,  -224, -416,
	  -215, -410,  -192, -394,  -160, -382,  -130, -382,  -130, -373,
	  -70, -350,  -53, -354,  -50, -370,  -27, -400,  -28, -418,
	  -24, -432,  -28, -440,  -20, -447,  -8, -475,  -16, -482,
	  -5, -477,  0, -500,  22, -500,  41, -515,  55, -526,  // Kourou
	  68, -550,  65, -573,  80, -600,  80, -610,  90, -615,
	  100, -620,  115, -614,  110, -640,  104, -648,  108, -660,
	  108, -677,  118, -677,  120, -685,  120, -693, 125, -693,
	  125, -705,  120, -705,  115, -720,  120, -723,  125, -717,
	  120, -727, 120, -740, 115, -740,  115, -750, 110, -758,
	  90, -760,  80, -775,  85, -780,  90, -790,  // Portobello Panama
	  // Antarctica   174  (cum 1262)
	  1111, 0,
	  -778, 1800,  -770, 1785,  -770, 1700,  -770, 1660,  -775, 1650,
	  -754, 1640,  -740, 1655,  -725, 1700,  -720, 1700,  -710, 1650,
	  -700, 1600,  -677, 1545,  -680, 1520,  -670, 1440,  -665, 1355,
	  -665, 1300,  -670, 1300,  -670, 1245,  -675, 1180,  -660, 1135,
	  -670, 1090,  -662, 1010,  -667, 1000,  -667, 900,  -675, 815,
	  -678, 780,  -694, 750,  -680, 700,  -676, 697,  -674, 600,
	  -662, 550,  -667, 500,  -675, 500,  -668, 485,  -675, 470,
	  -675, 455,  -687, 400,  -700, 390,  -680, 345,  -690, 330,
	  -705, 245,  -700, 135,  -700, 0,  -710, -90,  -718, -130,
	  -722, -125, -730, -190,  -740, -210,  -745, -250,  -758, -250,
	  -774, -320,  -776, -350,  -782, -355,  -780, -450,  -750, -610,
	  -728, -600,  -700, -615,  -667, -610,  -640, -560,  -630, -570,
	  -645, -635,  -640, -640,  -648, -640,  -660, -660,  -670, -670,
	  -698, -683,  -682, -705,  -690, -727,  -720, -750,  -728, -730,
	  -725, -790,  -738, -820,  -724, -950,  -720, -955,  -724, -1040,
	  -750, -1015,  -740, -1125,  -730, -1235,  -727, -1270,  -740, -1280,
	  -750, -1390,  -768, -1500,  -774, -1540,  -770, -1580,  -780, -1640,
	  -778, -1800,
	  // Europe West and South  [238]  (cum 1490)
	  1111, 0,          //  Atlantic and Mediterranean coast 132
	  540, 80,  530, 50,  525, 40,  515, 20,  502, 22,
	  492, -4,  497, -22,  480, -20,  480, -45,  475, -45,
	  470, -25,  465, -20,  435, -20,  436, -78,  430, -84,
	  416, -80,  380, -91,  377,  -85,  373, -85,  373, -75,
	  367, -58,  368, -52,  370, -45,  370, -23,  380, 0,
	  390, -5,  417, 17,  422, 30,  428, 30,  430, 50,
	  427, 65,  442, 85,  438, 103,  426, 106,  400, 150,
	  380, 160,  376, 155,  376, 161,  395, 170,  405, 173,
	  400, 176,  412, 170,   422, 167,  425, 140,  430, 130,
	  438, 125,  456, 125,  458, 130,  450, 135,  455, 140,
	  435, 155,  420, 190,  405, 190,  377, 210,  375, 210,
	  370, 218,  365, 227,  375, 225,  375, 230,  377, 240,
	  405, 225,  402, 230,  412, 250,  410, 268,  400, 264,  // Agean
	  1111, 0,         // Black Sea  48
	  410, 268,  415, 283,  415, 290,  418, 290,  422, 280,
	  427, 275,  465, 306,  460, 330,  455, 325,  444, 337,
	  455, 365,  460, 350,  470, 380,  455, 373,  450, 368,
	  425, 415,  420, 415,  413, 400,  416, 370,  422, 350,
	  421, 327,  415, 310,  415, 290,
	  1111, 0,        // Asia Minor  26
	  415, 283,  410, 295,  404, 264,  367, 273,  362, 300,
	  365, 307,  360, 325,  370, 358,  350, 355,  317, 345,
	  315, 330,  318, 320,
	  1111, 0,        // Caspian Sea   22
	  450, 470,  400, 500,  374, 490,  370, 540,  400, 540,
	  450, 505,  455, 525,  470, 525,  470, 500,  450, 470,
	  //  Europe North   160  (cum  1650)
	  1111, 0,   // 160
	  540, 80,  555, 80,  555, 75,  570, 76,  574, 105,
	  570, 102,  570, 106,  560, 100,  558, 107,  550, 100,
	  540, 110,  545, 135,  535, 145,  548, 178,  540, 178,
	  550, 200,  550, 210,  570, 210,  575, 222,  570, 226,
	  578, 240,  575, 220,  580, 220,  590, 230,  600, 285,
	  605, 280,  600, 223,  632, 220,  650, 250,  660, 240,
	  660, 220,  645, 220,  625, 175,  610, 175,  600, 182,
	  580, 170,  565, 160,  565, 145,  555, 145,  555, 128,
	  590, 107,  577, 75,  577, 66,  580, 55,  582, 60,
	  585, 50,  622, 50,  640, 100,  678, 158,  670, 125,
	  690, 160,  700, 180,  705, 220,  712, 220,  717, 255,
	  700, 250,  715, 275,  705, 310,  700, 295,  677, 405,
	  670, 410,  667, 404,  664, 377,  670, 340,  663, 350,
	  640, 350,  630, 370,  650, 365,  645, 400,  660, 400,
	  667, 423,  665, 440,  680, 440,  678, 456,  673, 465,
	  675, 450,  670, 456,  670, 472,  675, 476,   // indiga
	  //  Sjaeland  12
	  1111, 0,
	  565, 113,  548, 110,  545, 120, 566, 120, 565, 113,
	  //  Scicily   10
	  1111, 0,
	  375, 123,  365, 150,  376, 155,  375, 123,
	  //  Sardinia   14
	  1111, 0,
	  415, 76,  384, 77,  386, 95,  406, 95,  412, 90,
	  415, 76,
	  //  Corsica   12
	  1111, 0,
	  427, 90,  425, 78,  420, 80,  418, 85,  427, 90,
	  //  Cyprus  14
	  1111, 0,
	  352, 322,  346, 325,  350, 335,  357, 340,  355, 325,
	  352, 322,
	  //  Crete   12
	  1111, 0,
	  355, 230,  350, 265,  355, 265,  360, 230,  355, 230,
	  //  England   60   (cum 1784)
	  1111, 0,
	  500, -60,  505, -40,  510, -40,  510, 0,  515, 15,
	  518, 10,  527, 20,  525, 0,  528, 4,  565, -24,
	  565, -28,  577, -20,  576, -40,  580, -26,  580, -50,
	  575, -70,  576, -60,  573, -65,  560, -46,  548, -50,
	  550, -27,  542, -30,  530, -27,  530, -45,  525, -40,
	  520, -50,  520, -27,  516, -40,  500, -60,
	  //  Ireland   28
	  1111, 0,
	  540, -100,  530, -100,  530, -85,  522, -106,  518, -80,
	  524, -70,  527, -67,  532, -70,  540, -58,  555, -61,
	  555, -75,  542, -78,  540, -100,
	  //  Iceland  36   (cum 1848)
	  1111, 0,
	  636, -230,  630, -180,  640, -145,  658, -140,  667, -160,
	  660, -177,  665, -185,  660, -190,  665, -200,  654, -220,
	  665, -230,  657, -245,  655, -225,  650, -240,  645, -227,
	  640, -225, 636, -230,
	  //  Arabia  44
	  1111, 0,
	  297, 322,  275, 340,  290, 350,  278, 348,  170, 420,
	  125, 430,  122, 440,  160, 520,  165, 524,  170, 550,
	  190, 585,  225, 595,  250, 560,  240, 540,  240, 515,
	  295, 490,  300, 500,  260, 540,  270, 560,  250, 575,
	  250, 662,
	  //  South Asia  194  (cum 2086)
	  240, 675,  225, 700,  220, 690,  208, 705,  218, 725,
	  186, 725,  160, 735,  80, 775,  94, 785,  102, 790,
	  102, 797,  120, 797,  136, 805,  150, 800,  165, 810,
	  167, 823,  172, 823,  176, 830,  195, 853,  200, 862,
	  210, 870,  216, 868,  221, 920,  203, 926,  200, 930,
	  195, 937,  193, 933,  183, 940,  175, 945,  160, 940,
	  158, 954,  170, 970,  125, 984,  117, 980,  113, 985,
	  80, 980,  65, 1005,  45, 1005,  15, 1040,  15, 1045,
	  30, 1040,  60, 1040,  80, 1000,  90, 1000,  90, 990,
	  138, 1000,  138, 1010,  130, 1010,  130, 1020,  120, 1030,
	  108, 1030,  100, 1050,  87, 1050,  112, 1090,  155, 1085,
	  190, 1060,  205, 1070,  210, 1095,  200, 1100,  193, 1085,
	  180, 1085,  180, 1100,  200, 1105,  206, 1100,  215, 1135,
	  220, 1135,  215, 1140,  227, 1166,  250, 1195,  260, 1195,
	  278, 1207,  300, 1224,  302, 1205,  310, 1217,  345, 1202,
	  350, 1193,  365, 1200,  375, 1230,  375, 1205,  370, 1190,
	  375, 1190,  375, 1175,  390, 1175,  390, 1190,  407, 1215,
	  395, 1215,  400, 1240,  395, 1250,  378, 1247,  375, 1265,
	  344, 1265,  350, 1295,  375, 1290,  400, 1275,  407, 1300,
	  420, 1300,  427, 1323,       //  Vladivostok
	  //  Russian Pacific coastline   110  (cum 2196)
	  423, 1330,  430, 1345,  478, 1400,  523, 1410,  525, 1420,
	  530, 1415,  532, 1420,  542, 1395,  542, 1380,  535, 1380,
	  540, 1380,  535, 1370,  540, 1370,  548, 1352,  592, 1455,
	  595, 1500,  590, 1533,  584, 1520,  590, 1550,  598, 1545,
	  610, 1557,  620, 1570,  622, 1600,  608, 1600,  620, 1630,
	  625, 1630,  625, 1645,  610, 1632,  604, 1620,  575, 1575,
	  575, 1565,  570, 1565,  560, 1555,  510, 1567,  515, 1575,
	  520, 1580,  525, 1580,  525, 1600,  540, 1600,  545, 1623,
	  565, 1620,  565, 1630,  575, 1629,  575, 1624,  600, 1630,
	  606, 1660,  600, 1660,  606, 1670,  606, 1690,  600, 1700,
	  625, 1772,  623, 1792,  645, 1775,  645, 1780,  650, 1800,
	  //  Artic Asian coastline   160  (cum 2356)
	  1111, 0,
	  675, 476,  685, 545,  680, 540,  685, 590,  680, 600,
	  685, 610,  705, 580,  707, 590,  698, 605,  697, 640,
	  678, 670,  685, 680,  693, 670,  720, 670,  730, 695,
	  730, 725,  665, 725,  665, 740,  720, 740,  725, 750,
	  715, 750,  715, 770,  720, 760,  723, 780,  720, 825,
	  725, 810,  735, 810,  740, 860,  750, 805,  765, 950,
	  760, 970,  770, 1010,  775, 1010,  780, 1045,  770, 1060,
	  773, 1110,  760, 1140,  735, 1060,  740, 1130,  740, 1190,
	  735, 1190,  730, 1230,  737, 1230,  735, 1280,  715, 1290,
	  710, 1315,  720, 1325,  715, 1330,  720, 1400,  725, 1390,
	  727, 1410,  720, 1500,  715, 1520,  715, 1575,  705, 1600,
	  695, 1595,  695, 1670,  700, 1680,  685, 1690,  685, 1710,
	  700, 1705,  698, 1760,  682, 1800,
	  1111, 0,
	  682, -1800,  670, -1750,  665, -1745,  670, -1745,  670, -1720,
	  665, -1700,  660, -1710,  660, -1725,  640, -1725,  650, -1758,
	  660, -1760,  660, -1780,  665, -1780,  665, -1795,  650, -1800,
	  //  Lake Baykal    18
	  1111, 0,
	  520, 1030,  520, 1050,  555, 1090,  555, 1098,  535, 1090,
	  520, 1070,  515, 1040,  520, 1030,
	  //  Sakhalin       20
	  1111, 0,
	  535, 1420,  460, 1425,  465, 1430,  460, 1435,  475, 1430,
	  490, 1430,  490, 1440,  544, 1425,  535, 1420,
	  //  Ostrov Kotel'ny   18
	  1111, 0,
	  765, 1380,  753, 1375,  745, 1390,  750, 1440,  705, 1450,
	  767, 1410,  760, 1410,  765, 1380,
	  //  Svernaya Semlya    22
	  1111, 0,
	  810, 950,  800, 910,  785, 950,  785, 1000,  776, 1000,
	  780, 1050,  790, 1020,  785, 1000,  800, 1000,  810, 950,
	  //  Novaya Semlya     24   (cum 2458)
	  1111, 0,
	  770, 680,  765, 600,  750, 560,  720, 520,  710, 540,
	  710, 575,  720, 555,  725, 560,  750, 600,  765, 680,
	  770, 680,
	  //  Ceylon      14
	  1111, 0,
	  96, 800,  65, 800,  60, 810,  65, 820,  90, 810,  96, 800,
	  //  Taiwan    14
	  1111, 0,
	  235, 1200,  220, 1205,  220, 1210,  245, 1220,  240, 1215,
	  235, 1200,
	  //  Japan    76   (cum 2562)
	  1111, 0,    // Kyushu
	  340, 1308,  338, 1295,  308, 1300,  310, 1315,  330, 1320,
	  340, 1308,
	  1111, 0,    // Honshu
	  340, 1308,  345, 1350,  340, 1346,  335, 1353,  345, 1360,
	  350, 1405,  395, 1415,  410, 1410,  405, 1400,  390, 1395,
	  380, 1390,  370, 1360,  350, 1355,  350, 1330,  340, 1308,
	  1111, 0,    // Hokkaido
	  415, 1410,  410, 1400,  420, 1400,  425, 1415,  452, 1420,
	  430, 1450,  415, 1430,  420, 1410, 415, 1410,
	  1111, 0,    // Shikoku
	  335, 1325,  330, 1330,  335, 1340,  335, 1325,
	  //  Sumatra    26
	  1111, 0,
	  55, 950,  15, 990,  2, 990,  -60, 1040,  -60, 1050,
	  -30, 1050,  -28, 1045,  -10, 1045,  -10, 1035,  2, 1035,
	  50, 970,  55, 950,
	  //  Java       16
	  1111, 0,
	  -60, 1055,  -70, 1050,  -85, 1140,  -80, 1140,  -70, 1105,
	  -75, 1103,  -60, 1055,
	  //  Borneo    28
	  1111, 0,
	  15, 1090,  0, 1090,  -35, 1100,  -45, 1145,  -40, 1155,
	  5, 1180,  5, 1190,  40, 1180,  50, 1190,  68, 1170,
	  25, 1110,  15, 1110,  15, 1090,
	  //  Celebes    40
	  1111, 0,
	  7, 1200,  -40, 1190,  -40, 1195,  -60, 1195,  -60, 1204,
	  -35, 1205,  -35, 1210,  -50, 1215,  -50, 1220,  -20, 1210,
	  -5, 1230,  -15, 1210,  0, 1200,  5, 1205,  5, 1240,
	  15, 1250,  10, 1240,  10, 1207,  7, 1200,
	  //  New Guinea   52  (cum 2724)
	  1111, 0,
	  -10, 1310,  -25, 1338,  -30, 1320,  -40, 1330,  -40, 1335,
	  -35, 1340,  -60, 1370,  -75, 1375,  -80, 1367,  -80, 1400,
	  -90, 1404,  -90, 1430,  -70, 1435,  -70, 1455,  -100, 1470,
	  -105, 1504,  -60, 1465,  -60, 1470,  -50, 1450,  -45, 1450,
	  -18, 1380,  -35, 1350,  -8, 1340,  -2, 1320,  -10, 1310,
	  //  Timor    14
	  1111, 0,
	  -100, 1240,  -100, 1245,  -85, 1270,  -85, 1250,  -95, 1240,
	  -100, 1240,
	  //  Philippines   66   (cum 2804)
	  1111, 0,
	  190, 1204,  160, 1204,  162, 1200,  123, 1210,  140, 1215,
	  130, 1240,  140, 1240,  150, 1220,  155, 1215,  175, 1225,
	  190, 1225,  190, 1204,
	  1111, 0,
	  117, 1220,  103, 1220,  105, 1225,  100, 1220,  90, 1230,
	  110, 1230,  117, 1220,
	  1111, 0,
	  98, 1260,  80, 1240,  85, 1240,  75, 1220,  80, 1240,
	  60, 1240,  55, 1255,  70, 1255,  60, 1260,  83, 1265,
	  98, 1260,
	  //  Hawaii       34  (cum 2838)
	  1111, 0,
	  202, -1559,  197, -1561,  189, -1556,  195, -1548,  202, -1559,
	  1111, 0,
	  210, -1566,  206, -1554,  207, -1560,  210, -1556,
	  1111, 0,
	  216, -1583,  213, -1582,  212, -1577,  217, -1580,  216, -1583,
	  // Puerto Rico  14
	  1111, 0,
	  185, -672,  179, -672,  180, -658,  182, -656,  184, -656,
	  185, -672,
	  // Haiti    28
	  1111, 0,
	  185, -740,  180, -736,  182, -720,  176, -715,  183, -705,
	  182, -703,  184, -686,  186, -686,  197, -705,  194, -716,
	  197, -734,  185, -726,  185, -740,
	  // Cuba  24
	  1111, 0,
	  197, -776,  200, -740,  230, -797,  230, -825,  225, -845,
	  212, -845,  217, -820,  210, -790,  205, -780,  202, -766,
	  197, -776,
	  // Jamaica    12   (cum 2916)
	  1111, 0,
	  182, -785,  175, -770,  180, -765,  185, -780,  182, -785,
	  //  North America  326
	  1111, 0,
	  90, -796,  80, -800,  80, -820,  85, -830,  82, -845,
	  100, -852,  95, -850,  110, -860,  130, -870,  145, -920,
	  160, -955,  154, -965,  200, -1057,  215, -1050,  308, -1135,
	  310, -1140,  300, -1140,  235, -1100,  270, -1150,  270, -1145,
	  330, -1180,  335, -1200,  375, -1230,  375, -1225,  380, -1230,
	  403, -1250,  470, -1250,  480, -1255,  505, -1285,  515, -1285,
	  540, -1300,  545, -1340,  600, -1400,  600, -1450,  606, -1465,
	  606, -1496,  600, -1470,  594, -1525,  600, -1525,  582, -1520,
	  577, -1525,  570, -1520,  565, -1550,  570, -1555,  550, -1600,
	  535, -1655,  552, -1640,  575, -1580,  582, -1580,  582, -1620,
	  600, -1620,  600, -1650,  620, -1665,  640, -1650,  645, -1618,
	  645, -1665,  655, -1690,  667, -1640,  662, -1640,  662, -1630,
	  670, -1630,  687, -1665,  690, -1640,  700, -1630,  715, -1570,
	  700, -1460,  700, -1420,  685, -1355,  695, -1350,  702, -1290,
	  692, -1260,  700, -1250,  692, -1245,  697, -1235,  680, -1150,
	  675, -1155,  675, -1100,  667, -1080,  685, -1080,  675, -970,
	  684, -935,  700, -960,  745, -960,  745, -900,  730, -945,
	  685, -900,  690, -895,  670, -870,  670, -865,  700, -860,
	  710, -900,  740, -900,  740, -785,  700, -680,  695, -700,
	  667, -610,  640, -640,  667, -680,  630, -650,  620, -655,
	  650, -780,  670, -720,  700, -780,  700, -820,  670, -810,
	  660, -860,  640, -800,  640, -870,  650, -880,  610, -940,
	  570, -930,  545, -825,  513, -800,  520, -790,  545, -800,
	  555, -770,  630, -780,  630, -740,  610, -700,  588, -700,
	  588, -660,  600, -650,  520, -570,  500, -570,  470, -520,
	  480, -590,  515, -575,  500, -600,  500, -670,  460, -640,
	  450, -610,  437, -655,  450, -645,  430, -705,  410, -700,
	  402, -740,  360, -760,  350, -755,  310, -810,  300, -810,
	  280, -800,  250, -802,  250, -810,  282, -835,  300, -833,
	  300, -900,  290, -900,  290, -920,  298, -935,  290, -970,
	  260, -970,  210, -980,  195, -965,  195, -910,  210, -900,
	  212, -875,  160, -885,  165, -850,  150, -840,  100, -840,
	  95, -820,  90, -790,
	  //  Vancouver Island / Puget Sound     10
	  1111, 0,
	  485, -1255,  480, -1240,  500, -1250,  505, -1285,
	  //  Cook Inlet    6
	  1111, 0,
	  600, -1525,  628, -1500,
	  //  Baffin Island Sounds   28  (cum 3286)
	  1111, 0,
	  700, -860,  700, -820,
	  1111, 0,
	  740, -850,  720, -850, 710, -845,
	  1111, 0,
	  740, -803,  730, -803,  730, -760,
	  1111, 0,
	  630, -698,  625, -655,
	  //  Great Lakes   76  (cum 3362)
	  1111, 0,
	  468, -920,  490, -880,  490, -850,  465, -845,  468, -920,
	  1111, 0,
	  459, -870,  430, -880,  415, -875,  420, -865,  440, -865,
	  460, -850, 459, -870,
	  1111, 0,
	  457, -847,  455, -835,  435, -835,  440, -830,  430, -825,
	  432, -818,  450, -815,  445, -800,  460, -810,  462, -838,
	  457, -847,
	  1111, 0,
	  420, -835,  415, -838,  413, -820,  427, -790,  420, -835,
	  1111, 0,
	  432, -798,  435, -763,  440, -765,  438, -790,  432, -798,
	  //  Greenland   78  (cum 3440)
	  1111, 0,
	  765, -700,  757, -590,  720, -560,  715, -520,  710, -520,
	  715, -540,  690, -535,  695, -510,  677, -510,  675, -530,
	  650, -530,  625, -495,  600, -430,  650, -400,  670, -330,
	  680, -250,  700, -220,  705, -250,  715, -250,  705, -230,
	  705, -210,  730, -220,  735, -200,  795, -200,  795, -180,
	  820, -102,  820, -300,  825, -200,  840, -350,  825, -500,
	  820, -600,  800, -700,  800, -650,  790, -660,  780, -730,
	  775, -680,  770, -705,  765, -700,
	  //   Ellesmere Island   58
	  1111, 0,
	  800, -970,  820, -940,  830, -750,  830, -700,  825, -610,
	  795, -700,  765, -770,  760, -900,  755, -895,  760, -810,
	  745, -800,  745, -920,  770, -950,  770, -900,  775, -880,
	  770, -850,  780, -880,  775, -910,  800, -970,
	  1111, 0,
	  820, -920,  810, -880,  810, -840,  815, -800,
	  1111, 0,
	  780, -870,  785, -850,  805, -880,
	  //   Melville Island    32
	  1111, 0,
	  760, -1220,  745, -1130,  750, -1060,  765, -1070,  765, -1080,
	  770, -1100,  755, -1090,  755, -1110,  770, -1165,  775, -1175,
	  760, -1220,
	  1111, 0,
	  755, -1180,  770, -1165,  760, -1200,
	  //   Victoria Island   68   (cum 3598)
	  1111, 0,
	  720, -1260,  710, -1235,  715, -1205,  723, -1200,  725, -1180,
	  720, -1180,  720, -1160,  715, -1180,  703, -1110,  700, -1170,
	  690, -1140,  675, -1140,  680, -1090,  690, -1070,  680, -1020,
	  695, -1020,  695, -1008,  715, -1050,  740, -1050,  740, -1070,
	  720, -1070,  720, -1075,  730, -1100,  725, -1100,  735, -1150,
	  740, -1180,  737, -1200,  745, -1215,  745, -1255,  720, -1260,
	  1111, 0,
	  725, -1190,  735, -1150,

	  };

  if ((mapprj == 3) && (physmap))
   {
    if(centmer == 0) WorldMap.drawPixmap(0,0,*buffPWM);
    else
     {
      ofz = int(centmer);
      if (ofz < -180) ofz = -180;
      if (ofz > 360) ofz = 0;
      if (ofz < 0) ofz += 360;
      ofz = 2*ofz;  
      WorldMap.drawPixmap(-ofz,0,*buffPWM);
      ofz = 720 - ofz;
      WorldMap.drawPixmap(ofz,0,*buffPWM);
     };
   }
  else
   {
    WorldMap.setPen(QColor( 200, 100, 50 ));
    k = gx;
    lon = 0;
    for (j=0; j<k; j+=2)
     {
      if (geo[j] == 1111) lon = 0;
      else
       {
        lat = double(geo[j]) * 0.1;
        lng = double(geo[j+1]) * 0.1;
        if (!lon) p1 = PxCoord(lat,lng);
        else
         {
          p2 = PxCoord(lat,lng);
          if (abs((p1.x() - p2.x())) < 200)
            WorldMap.drawLine(p1, p2);
          p1 = p2;
         };  
        lon = 1;
       };
     };
   };

  // mark equator for stereographic projection
  if ((mapprj == 1) || (mapprj == 2))
   {
    WorldMap.setPen(QColor( 130, 160, 180 ));
    lat = 0;
    k = 0;
    p1 = PxCoord(0, 0.0001);
    for (j=0; j<360; j++)
     {
      lng = double(j);
      p2 = PxCoord(lat, lng);
      if ((p2.x() > mapwidth)
               || (p2.y() > mapheight)) k = 0;
      else
       {
        if (k) WorldMap.drawLine(p1, p2);
        k = 1;
        p1 = p2;
       };
     };
   };  // mapprj

 }
