/***************************************************************************
                          luncalview.h  -  description
                             -------------------
    Open Source Code. See main.cpp for details.
    copyright            : (C) 2011 by Holtkamp
 ***************************************************************************/

#ifndef LUNACALVIEW_H
#define LUNACALVIEW_H

#include <QWidget>
#include<QScrollArea>
#include<QScrollBar>
#include <QPixmap>
#include <QPoint>
#include <QMouseEvent>
#include <QPaintEvent>

#include "lunamph.h"

static const int GBL_MAPWIDTH = 721;   // width of map in pixel
static const int GBL_MAPHEIGHT = 361;  // heigth of map in pixel
static const int GBL_MAPTEXTHT = 50;  // height of text part of the map
static const int GBL_TEXTWIDTH = 860;
static const int GBL_TEXTLENGTH = 2500; // length of text buffer

class LunacalView : public QScrollArea
{
  Q_OBJECT
  public:
    LunacalView(QWidget *parent);
    ~LunacalView();
    void showtxt(bool tab, WoutData* outdata);  // output text on screen
    void showEcltxt(WinpData* indata); // output eclipse text on screen and prepare detailed eclipse
    void showEclDetails(int y);
    void showEclMap();
    void showSatEcl(WinpData* indata);  // special (satellite) eclipse map.
    void setMap(int mgrid, int mgrey, int mstep, int mapprj, double centmer); // set eclipse display parameters
    void getMapPos(int x, int y);
    void setFont(int fnt);
    void keyPress(int n);
    void scrollDsp(int dx, int dy, WinpData* indata);

    void pltCurrent(WinpData* indata);  // plot current step (special eclipse)
    void pltStep(WinpData* indata);   // plot next step (special eclipse)
    void pltBack(WinpData* indata);   // plot backward step (special eclipse)
    void PlotWorld(QPainter &WorldMap, WinpData* indata);  // Plot the coordinate grid and continents
    QPoint PxCoord (double lat, double lng); // get pixel coordinates
    void CoordPx (QPoint& p, double& lat, double& lng); // get coordinates from pixel
    void PltEclBound (QPainter &WorldMap, WinpData* indata);  // plot eclipse boundaries
    void PltCenLin (QPainter &WorldMap, WinpData* indata);  // plot center line
    void PltMaxPos (QPainter &WorldMap, WinpData* indata);  // plot position of maximum eclipse
    void MapLand(QPainter &WorldMap);   // draw the continents and major islands

    int origx;      // x pixel of coordinate origin
    int origy;      // y pixel of coordinate origin
    double xscale;  // scale in x (pixel / degrees)
    double yscale;  // scale in y (pixel / degrees)
  
  protected:
    virtual void paintEvent( QPaintEvent *e );
    virtual void mousePressEvent( QMouseEvent *e );
    virtual void update();

  private:
    QPixmap *buffer;
    QPixmap *buffoot;
    QPixmap *buffPWM; // physical world map 
    QPixmap *satbuf;  // buffer for SatEcl map
};

#endif
