#if !defined(__lunamph_h)
#define __lunamph_h

#include "attlib.h"

const int GBL_MAXLIN = 110;  // maximum number of lines in Window Buffer
const int GBL_MAXCPL = 77;  // maximum number of characters per line in buffer
const int GBL_ECLBUF = 12;  // maximum number of eclipses allowed

struct EclBuf                 // Buffer for Eclipse dates
 {
  int numecl;   // number of eclipses of the year;
  int eclselect; // eclipse selected for detailed calculation
  int lastyear;   // last year for which moon phases were calculated
  double lasttz;  // last used timezone
  double lastdlt; // last delta-t used
  int cstep;   // step width in minutes used for central eclipse output
  int ecloutbn; // output batch number of the detailed eclipse output
  double eclmjd [GBL_ECLBUF];  // the MJD's of the middle of the eclipses
  double magnitude [GBL_ECLBUF];   // magnitude of respective eclipse
  int phase [GBL_ECLBUF];    // phase of the eclipse
  int lccnt; // = 0 not visible, 1,2 = first rise/set; 3,4 second r/s
  double lcb1, lcb2, lce1, lce2;  // rise and set times (up to 2 possible)
  double jdmaxps; // MJD of maximum phase
  double maxps;  // value of maximum phase
  double maxelv; // elevation at maximum phase
  double ltotb;  // begin of totality/annularity at local position
  double ltote;  // end of totality/annularity at local position
  int emagrid;  // 0 if no grid, 1 if 30 deg grid, 2 if 10 deg grid
  int emagrey;  // 0 if map background white, 1 if grey
  int cphs;   // phase of the central eclipse for center line plotting
  double clat; // current latitude of center line (in decimal degrees)
  double clng; // current longitude of center line (in decimal degrees)
  double cmxlat; // latitude of maximum eclipse (in decimal degrees)
  double cmxlng; // longitude of maximum eclipse (in decimal degrees) 
  double jdstart; // (global) start jd of eclipse
  double jdstop;  // (global) end jd of eclipse
  };

struct WinpData   // Input from and data exchange with Main Window
 {
  int day;           // date
  int month;
  int year;
  int hour;
  int minute;
  int second;
  double mjd;        // current MJD used for special eclipse
  double tstep;      // time step (in sec) used for special eclipse
  double time;
  double tzone;      // timezone in hours
  double del_tdut;   // TDT - UT in sec
  double spnaz;      // azimuth of solar panel
  double spnel;      // elevation of solar panel
  double spneff;     // efficiency of solar panel
  int spnm1;         // beginning month of solar panel interval 
  int spnm2;         // end month of solar panel interval
  int del_auto;      // 1 = automatic del_tdut,  0 = manual
  int tmfirst;       // 1 = first call to sky or day details
  int svfapnd;       // 1 = append consecutive saves, 0 = overwrite
  int svfok;         // 1 = file can be written to, 0 = do not write on file
  int svfirst;       // 1 = first call to this save file, 0 = consecutive calls
  int invalflg;      // 1 = if special invalidation necessary
  double geolat;     // geographic coordinates (angles in degrees)
  double geolong;
  double geoheight;  // in meters
  double eclheight;  // height above Earth surface in km for special eclipse
  bool geoid;        // true if special eclipse surface ellipsoid; sphere else.
  char   geoname[31];
  char   svfname[55]; // name of save file
  int dmsflg;     // 0 = decimal degrees, 1 = degrees, minutes, seconds
  int filsflg;  //  1, 2 = if file save active
  int lstcall;  // 1 = last call moonph; 2 = caldisp; 0 = no call 
  int locecl;   // 1 = local eclipse call; 0 = normal (global) call
  int mapprj;     // 0 = world map, 1 = North stereogr., 2 = South st., 3 = physical
  double centmer;   // central meridian of the standard world map in deg.
  EclBuf eclbuf;
 };

struct WoutData                // Main Window output text buffer
 {
  int curlines;  // currently active lines
  char wbuf[GBL_MAXLIN][GBL_MAXCPL];  // output buffer
  double wsolstice; // MJD of winter solstice 
  };

void moonph(WinpData* indata, WoutData* outdata);  // calculate phases of the Moon
void caldisp(WinpData* indata, WoutData* outdata); // display calendar
void GetEclTxt (WinpData* indata, int j, char* jtxt);// get text for j-th ecl.
void PutEclLst (WinpData* indata, WoutData* outdata); // display list of Eclipses
void EclDetails(WinpData* indata,char *otxt);
void SatEcl(double jd, double tdut, double height, bool geoid, int &et1, double &lat1,
                       double &lng1, int &et2, double &lat2, double &lng2);
void EclPltCen(WinpData* indata, char *otxt);
void GetMaxPos(WinpData* indata); // get position of maximum eclipse
void InitBound(WinpData* indata, Vec3& ubm, Vec3& ube, Vec3& udm, Vec3& ude,
                                 Vec3& lbe, Vec3& lde);
void GNSBound(WinpData* indata, double t, Vec3& ubm, Vec3& ube, Vec3& udm,  
                        Vec3& ude, double& lat, double& lng);
void InRSBound(WinpData* indata, Vec3& ubm, Vec3& ube, Vec3& udm, Vec3& ude,
                                       double& dpb, double& dpd);
void GRSBound(WinpData* indata, double t, Vec3& ubm, Vec3& ube, Vec3& udm,
                                Vec3& ude, double dpb, double dpd,
                                double& lat1, double& lng1, double& lat2, double& lng2);
double DegFDms (double h);  // convert from Deg.fff into Deg.MMSSS
double DmsDegF (double h);  // convert from Deg.MMSSS into Deg.fff
void GetEaster (int yr, char* otxt); // Easter Holidays
void GetIslam (int yr, char* otxt);   // get Islamic Holidays
void GetIndia (int yr, char* otxt);   // get Indian New Year
void GetChina (WinpData* indata, WoutData* outdata, char *otxt);
void GetJewish (WinpData* indata, char* otxt);  // get Jewish holidays
void CompDayDet (WinpData* indata, WoutData* outdata, char* otxt, int k); 
void CompSkyDet (WinpData* indata, char* otxt, int k); // compute details of sky
void CompYearTot (WinpData* indata, double cang, char* otxt);  // yearly totals
void lwelcm(WinpData* indata, WoutData* outdata);  // welcoming message
void lweinp(WinpData* indata, WoutData* outdata);  // display initial input
int getcfg (WinpData* indata, int& swmax);  // get cfg from file
void putcfg (WinpData* indata, int swmax); // write configuration file
void DefTime (WinpData* indata); // Get System Time and Date
void dtmstr(double jdmoon, char *dts);  // get time string from julian date
void GetMonth (int mm, char* mchr);    // get month string from month number

#endif         // __lunamph_h sentry.

